import numpy


class Matrix:
    """
    This Matrix class represents a real 2D-matrix.
    """
    __data__: numpy.ndarray
    __shape__: (int, int)

    def __init__(self, data=None, shape=None, structure=None, model=None, n=None):
        """
        Creates a new matrix.
        Thy type of the matrix depends on the signature and arguments.

        - ``Matrix(numpy.ndarray)``: will create a new matrix with the given data in ndarray and its shape.
        - ``Matrix(list, (int,int))``: will create a new nxm matrix with the given rows and columns and data in list.
        - ``Matrix(list, str, int)``: will create a new square matrix of given size and structure of either \"unity\", \"diagonal\" or \"tridiagonal\"
        - ``Matrix(str, int)``: will create a new square matrix of given size and TODO


        :param data: Either a list or an numpy ndarray
        :param shape: A tuple containing the amount of rows and columns
        :param structure: Either \"unity\", \"diagonal\" or \"tridiagonal\"
        :param model: TODO
        :param n: Amount of rows of a square matrix or offset in case of diagonal structure

        :type data: list | numpy.ndarray
        :type shape: (int, int)
        :type structure: str
        :type model: str
        :type n: int

        :rtype: Matrix
        """
        if isinstance(data, numpy.ndarray):
            try:
                data.shape[1]
            except IndexError:
                self.__shape__ = (data.shape[0], 1)
            else:
                self.__shape__ = data.shape
            self.__data__ = data
        elif isinstance(data, list) and isinstance(shape, tuple):
            self.__shape__ = shape
            self.__data__ = numpy.array(data).reshape(shape)
        elif isinstance(data, list) and isinstance(structure, str) and isinstance(n, int):
            if structure == "unity":
                ...  # TODO: what does it mean?
            elif structure == "diagonal":
                diag = numpy.diag(data, n)
                self.__data__ = diag
                self.__shape__ = diag.shape
            elif structure == "tridiagonal":
                if len(data) != 3:
                    raise ValueError("If structure is tridiagonal, then the given data must be of length 3")
                tridiag = numpy.diag([data[0]] * (n-1), -1) + numpy.diag([data[1]] * n, 0) + numpy.diag([data[2]] * (n-1), 1)
                self.__data__ = tridiag
                self.__shape__ = tridiag.shape
        elif isinstance(model, str) and isinstance(n, int):
            ...  # TODO: what shall one do here?
        else:
            raise ValueError("Only following signatures are allowed: (numpy.ndarray), (list, tuple), (list, str, int), (str, int)")

    def get_data(self):
        """
        :return: the data of the matrix as a ``numpy.ndarray``
        """
        return self.__data__

    def shape(self):
        """
        :return: the shape of the matrix, which is ``(rows, columns)``
        """
        return self.__shape__

    def transpose(self):
        """
        :return: the transpose of the matrix
        """
        return Matrix(self.__data__.transpose())

    def __eq__(self, other):
        """
        Return ``self==value``

        :param other: The object to compare to; must be either a ``Matrix``, a ``list`` or a ``numpy.ndarray``
        :return: True if data in the matrix is totally equal to the given data in other, otherwise False
        """
        if isinstance(other, Matrix):
            return (self.__data__ == other.__data__).all()
        elif isinstance(other, list):
            return (self.__data__ == numpy.array(other)).all()
        elif isinstance(other, numpy.ndarray):
            return (self.__data__ == other).all()
        else:
            raise ValueError("Matrix type is not comparable to type of given ``other``")

    def __str__(self):
        return str(self.__data__)
